# cython: infer_types=True
# cython: cdivision=True
cimport cython
from os import path
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.exc cimport PyErr_CheckSignals
from libc.stdio cimport FILE, fopen, fclose, fread, fwrite, feof, fseek
from libc.errno cimport errno
from libc.string cimport memcpy

from libc.stdlib cimport qsort
from libc.stdint cimport int32_t

from cymem.cymem cimport Pool
from preshed.maps cimport PreshMap, MapStruct, map_get

from .sparse cimport SparseArray

from ..extra.eg cimport Example
from ..structs cimport SparseArrayC, SparseAverageC
from ..typedefs cimport class_t, count_t, time_t, feat_t
from ..linalg cimport Vec, VecVec
from .serialize cimport Writer
from .serialize cimport Reader


cdef class AveragedPerceptron:
    '''A linear model for online supervised classification.
    Expected use is via Cython --- the Python API is impoverished and inefficient.

    Emphasis is on efficiency for multi-class classification, where the number
    of classes is in the dozens or low hundreds.
    '''
    def __init__(self, templates, *args, **kwargs):
        self.weights = PreshMap()
        self.time = 0
        self.averages = PreshMap()
        self.mem = Pool()
        self.extracter = ConjunctionExtracter(templates)

    def __dealloc__(self):
        cdef size_t feat_addr
        # Use 'raw' memory management, instead of cymem.Pool, for weights.
        # The memory overhead of cymem becomes significant here.
        if self.weights is not None:
            for feat_addr in self.weights.values():
                if feat_addr != 0:
                    PyMem_Free(<SparseArrayC*>feat_addr)
        if self.averages is not None:
            for feat_addr in self.averages.values():
                if feat_addr != 0:
                    feat = <SparseAverageC*>feat_addr
                    PyMem_Free(feat.avgs)
                    PyMem_Free(feat.times)

    def __call__(self, Example eg):
        self.set_scoresC(eg.c.scores,
            eg.c.features, eg.c.nr_feat)
        PyErr_CheckSignals()
        return eg.guess

    def train_example(self, Example eg):
        self(eg)
        self.update(eg)
        return eg

    def predict_example(self, Example eg):
        self(eg)
        return eg

    def update(self, Example eg):
        self.updateC(&eg.c)

    def dump(self, loc):
        cdef Writer writer = Writer(loc, self.weights.length)
        cdef feat_t key
        cdef size_t feat_addr
        for i, (key, feat_addr) in enumerate(self.weights.items()):
            if feat_addr != 0:
                writer.write(key, <SparseArrayC*>feat_addr)
            if i % 1000 == 0:
                PyErr_CheckSignals()
        writer.close()

    def load(self, loc):
        cdef feat_t feat_id
        cdef SparseArrayC* feature
        cdef Reader reader = Reader(loc)
        self.weights = PreshMap(reader.nr_feat)
        cdef int i = 0
        while reader.read(self.mem, &feat_id, &feature):
            self.weights.set(feat_id, feature)
            if i % 1000 == 0:
                PyErr_CheckSignals()
            i += 1

    def end_training(self):
        cdef feat_id
        cdef size_t feat_addr
        cdef int i = 0
        for feat_id, feat_addr in self.averages.items():
            if feat_addr != 0:
                feat = <SparseAverageC*>feat_addr
                i = 0
                while feat.curr[i].key >= 0:
                    unchanged = (self.time + 1) - <time_t>feat.times[i].val
                    feat.avgs[i].val += unchanged * feat.curr[i].val
                    feat.curr[i].val = feat.avgs[i].val / self.time
                    i += 1

    property nr_feat:
        def __get__(self):
            return self.extracter.nr_templ

    cdef void set_scoresC(self, weight_t* scores, const FeatureC* feats, int nr_feat) nogil:
        # This is the main bottle-neck of spaCy --- where we spend all our time.
        # Typical sizes for the dependency parser model:
        # * weights_table: ~9 million entries
        # * n_feats: ~200
        # * scores: ~80 classes
        # 
        # I think the bottle-neck is actually reading the weights from main memory.
        cdef const MapStruct* weights_table = self.weights.c_map
        cdef int i, j
        cdef FeatureC feat
        for i in range(nr_feat):
            feat = feats[i]
            class_weights = <const SparseArrayC*>map_get(weights_table, feat.key)
            if class_weights != NULL:
                j = 0
                while class_weights[j].key >= 0:
                    scores[class_weights[j].key] += class_weights[j].val * feat.value
                    j += 1

    @cython.cdivision(True)
    cdef int updateC(self, const ExampleC* eg) except -1:
        self.time += 1
        guess = VecVec.arg_max_if_true(eg.scores, eg.is_valid, eg.nr_class)
        if eg.costs[guess] > 0:
            best = VecVec.arg_max_if_zero(eg.scores, eg.costs, eg.nr_class)
            for feat in eg.features[:eg.nr_feat]:
                self.update_weight(feat.key, best,   feat.value * eg.costs[guess])
                self.update_weight(feat.key, guess, -feat.value * eg.costs[guess])

    cpdef int update_weight(self, feat_t feat_id, class_t clas, weight_t upd) except -1:
        if upd == 0:
            return 0
        feat = <SparseAverageC*>self.averages.get(feat_id)
        if feat == NULL:
            feat = <SparseAverageC*>PyMem_Malloc(sizeof(SparseAverageC))
            if feat is NULL:
                msg = (feat_id, clas, upd)
                raise MemoryError("Error allocating memory for feature: %s" % msg)
            feat.curr  = SparseArray.init(clas, upd)
            feat.avgs  = SparseArray.init(clas, 0)
            feat.times = SparseArray.init(clas, <weight_t>self.time)
            self.averages.set(feat_id, feat)
            self.weights.set(feat_id, feat.curr)
        else:  
            i = SparseArray.find_key(feat.curr, clas)
            if i < 0:
                feat.curr = SparseArray.resize(feat.curr)
                feat.avgs = SparseArray.resize(feat.avgs)
                feat.times = SparseArray.resize(feat.times)
                self.weights.set(feat_id, feat.curr)
                i = SparseArray.find_key(feat.curr, clas)
            feat.curr[i].key = clas
            feat.avgs[i].key = clas
            # Apply the last round of updates, multiplied by the time unchanged
            feat.avgs[i].val += (self.time - feat.times[i].val) * feat.curr[i].val
            feat.curr[i].val += upd
            feat.times[i].val = self.time



from ..structs cimport FeatureC
from ..structs cimport ConstantsC

from ..typedefs cimport len_t
from ..typedefs cimport idx_t


cdef void d_ELU__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil
   

cdef void d_ReLu__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil


cdef void d_ELU__dot__normalize__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil


cdef void d_log_loss(
    float* loss,
        const float* costs,
        const float* scores,
            len_t nr_out) nogil


cdef void d_dot(float* btm_diff,
        int nr_btm,
        const float* top_diff, int nr_top,
        const float* W) nogil
 

cdef void d_ELU(float* delta, const float* signal_out, int n) nogil



from ..solve cimport *
from ..forward cimport *
from ..backward cimport *


def call_dot_plus(
    float[:] out,
        float[:] in_,
        float[:] W,
        float[:] bias,
        int nr_top,
        int nr_btm):
    dot_plus(&out[0],
        &bias[0], nr_top, &in_[0], nr_btm, &W[0])
    return out


def call_d_dot(
    float[:] btm_diff,
        float[:] top_diff,
        float[:] W,
        int nr_top,
        int nr_btm):
    d_dot(&btm_diff[0],
        nr_btm, &top_diff[0], nr_top, &W[0])
    return btm_diff


def call_ELU(float[:] out, int nr_out):
    ELU(&out[0], nr_out)
    return out


def call_d_ELU(float[:] delta, float[:] signal_out, int nr_out):
    d_ELU(&delta[0], &signal_out[0], nr_out)
    return delta



from cymem.cymem cimport Pool
from libc.math cimport sqrt as c_sqrt
from libc.string cimport memset, memcpy, memmove

from preshed.maps cimport map_init as Map_init
from preshed.maps cimport map_set as Map_set
from preshed.maps cimport map_get as Map_get

from ..structs cimport ExampleC, FeatureC, MapC
from ..typedefs cimport feat_t, weight_t, atom_t
from ..linalg cimport Vec, VecVec


cdef class Example:
    cdef Pool mem
    cdef ExampleC c

    cpdef int fill_atoms(self, atom_t value, int nr_atom) except -1
    cpdef int fill_features(self, int value, int nr_feat) except -1
    cpdef int fill_is_valid(self, int value, int nr_class) except -1
    cpdef int fill_costs(self, weight_t value, int nr_class) except -1
    cpdef int fill_scores(self, weight_t value, int nr_class) except -1

    @staticmethod
    cdef inline Example from_ptr(Pool mem, ExampleC* ptr):
        cdef Example eg = Example.__new__(Example)
        eg.mem = mem
        eg.c = ptr[0]
        return eg

    @staticmethod
    cdef inline void init(ExampleC* self, Pool mem, model_shape,
                          blocks_per_layer) except *:
        raise NotImplementedError
        # Revise this/make consistent with the Python-land reset_x() stuff
        #if len(model_shape) >= 2:
        #    self.nr_atom = model_shape[0]
        #    self.atoms = <atom_t*>mem.alloc(self.nr_atom, sizeof(self.atoms[0]))
        #self.fwd_state = <weight_t**>mem.alloc(len(model_shape), sizeof(void*))
        #self.bwd_state = <weight_t**>mem.alloc(len(model_shape), sizeof(void*))
        #self.widths = <int*>mem.alloc(len(model_shape), sizeof(int))
        #for i, width in enumerate(model_shape):
        #    self.widths[i] = width
        #    self.fwd_state[i] = <weight_t*>mem.alloc(width * blocks_per_layer,
        #                                             sizeof(weight_t))
        #    self.bwd_state[i] = <weight_t*>mem.alloc(width * blocks_per_layer,
        #                                             sizeof(weight_t))
        #self.nr_layer = len(model_shape)

        #self.nr_class = model_shape[-1]
        #self.scores = <weight_t*>mem.alloc(self.nr_class, sizeof(self.scores[0]))
        #self.is_valid = <int*>mem.alloc(self.nr_class, sizeof(self.is_valid[0]))
        #for i in range(self.nr_class):
        #    self.is_valid[i] = 1
        #self.costs = <weight_t*>mem.alloc(self.nr_class, sizeof(self.costs[0]))



# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
cimport cython

from ..typedefs cimport len_t
from ..typedefs cimport idx_t

from ..linalg cimport MatMat, MatVec, VecVec, Vec, sqrtf, expf


DEF EPS = 0.00000001 
DEF ALPHA = 1.0


cdef void d_ELU__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil:
    d_ELU(bwd[1],
        fwd[1], shape[1])
    # Set the gradient for F(W * fwd[0]) 
    MatMat.add_outer_i(gradient,
        bwd[1], fwd[0], shape[1], shape[0])
    VecVec.add_i(gradient + shape[1] * shape[0],
        bwd[1], 1.0, shape[1])
    # Set the partial derivative for bwd[0], so next step can set its gradient
    MatVec.T_dot(bwd[0],
        W, bwd[1], shape[1], shape[0])


cdef void d_ReLu__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil:
    d_ReLu(bwd[1],
        fwd[1], shape[1])
    # Set the gradient for F(W * fwd[0]) 
    MatMat.add_outer_i(gradient,
        bwd[1], fwd[0], shape[1], shape[0])
    VecVec.add_i(gradient + shape[1] * shape[0],
        bwd[1], 1.0, shape[1])
    # Set the partial derivative for bwd[0], so next step can set its gradient
    MatVec.T_dot(bwd[0],
        W, bwd[1], shape[1], shape[0])


cdef void d_ELU__dot__normalize__dot(float* gradient, float** bwd, float* averages,
        const float* W, const float* const* fwd, const len_t* shape,
        int nr_above, int nr_below, const ConstantsC* hp) nogil:
    # D{ELU(BN(Lin(x)))} = ELU'(BN(Lin(x))) * BN'(Lin(x)) * Lin'(x)
    d_ELU(bwd[1],
        fwd[1], shape[1])
    # At this point we have what the paper refers to as dE/dY. Set the gradient
    # for the bias and gamma params now.
    bias = W + (shape[1] * shape[0])
    gamma = bias + shape[1]
    x_norm = fwd[1] + shape[1]
    gamma_grad = gradient + (shape[1] * shape[0]) + shape[1]
    for i in range(shape[1]):
        gamma_grad[i] += bwd[1][i] * x_norm[i]
    VecVec.add_i(gradient + (shape[1] * shape[0]),
        bwd[1], 1.0, shape[1])
    # Now continue computing BN'. We transform dE/dY into dE/dX'.
    # We have to transform it into dE/dX, so that we can calculate the gradient
    # for our weights.
    VecVec.mul_i(bwd[1],
        gamma, shape[1])
    # Read the E(x), Var(x), E_dXh, E_dXh_dot_Xh estimates from 'averages'
    cdef const float* Ex = averages
    cdef const float* Vx = averages + shape[1]
    cdef float* E_dXh = averages + shape[1] * 2
    cdef float* E_dXh_Xh = averages + shape[1] * 3
    d_normalize(bwd[1], E_dXh, E_dXh_Xh,
        x_norm, Vx, shape[1], hp.a, hp.t)
    # Finally we have dE/dX. Now we can calculate the gradient of W
    MatMat.add_outer_i(gradient,
        bwd[1], fwd[0], shape[1], shape[0])
    # And calculate the error w.r.t the previous layer
    MatVec.T_dot(bwd[0],
        W, bwd[1], shape[1], shape[0])


cdef void d_dot(float* btm_diff,
        int nr_btm,
        const float* top_diff, int nr_top,
        const float* W) nogil:
    # And calculate the error w.r.t the previous layer
    MatVec.T_dot(btm_diff,
        W, top_diff, nr_top, nr_btm)
 

cdef void d_ELU(float* delta, const float* signal_out, int n) nogil:
    # Backprop the ELU transformation
    # Note that this is over the function _output_, not the function
    # _input_!
    for i in range(n):
        if signal_out[i] <= 0:
            delta[i] *= signal_out[i] + ALPHA



cdef void d_ReLu(float* delta, const float* signal_out, int n) nogil:
    # Backprop the ELU transformation
    # Note that this is over the function _output_, not the function
    # _input_!
    for i in range(n):
        if signal_out[i] <= 0:
            delta[i] = 0


cdef void d_log_loss(
    float* loss,
        const float* costs,
        const float* scores,
            len_t nr_out
) nogil:
    # This assumes only one true class
    cdef idx_t i
    for i in range(nr_out):
        loss[i] = scores[i] - (costs[i] == 0)


cdef void d_normalize(float* bwd, float* E_dEdXh, float* E_dEdXh_dot_Xh,
        const float* Xh, const float* Vx, len_t n, float alpha, float time) nogil:
    # Update EMA estimate of mean(dL/dX_hat)
    Vec.mul_i(E_dEdXh,
        alpha, n)
    VecVec.add_i(E_dEdXh,
        bwd, 1-alpha, n)
    # Update EMA estimate of mean(dE/dX_hat \cdot X_hat)
    Vec.mul_i(E_dEdXh_dot_Xh,
        alpha, n)
    for i in range(n):
        E_dEdXh_dot_Xh[i] += (1-alpha) * bwd[i] * Xh[i]
    # Simplification taken from Caffe, I think by cdoersch
    # if X' = (X-mean(X))/sqrt(var(X)+eps), then
    # dE/dX =
    #   (dE/dXh - mean(dE/dXh) - mean(dE/dXh * Xh) * Xh)
    #     ./ sqrt(var(X) + eps)
    # bwd is dE/dXh to start with. We change it to dE/dX in-place.
    if time >= 100:
        for i in range(n):
            bwd[i] -= E_dEdXh[i] - E_dEdXh_dot_Xh[i] * Xh[i]
            bwd[i] /= sqrtf(Vx[i] + EPS)
